"""Module for handling analytics models."""
import copy
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Union

import numpy as np
from numpy.typing import NDArray

from feltlabs.core import randomness
from feltlabs.core.models.base_model import AvgModel, BaseModel


@dataclass
class Metric:
    """Class for defining different analytics which can be calculated.

    Attributes:
        scale_rand: set true if random values are scaled by sample size
        fit_fn: function for calculating metric on original data
        agg_fun: aggregation function for combining results from different models
        remove_fn: function used for calculate negative of aggregated noise model
        output_fn: function used to calculate final value
    """

    scale_rand: bool
    fit_fn: Callable
    agg_fn: Callable
    remove_fn: Callable
    output_fn: Callable = lambda x, _: x


SUPPORTED_MODELS = {
    "Sum": Metric(
        scale_rand=True,
        fit_fn=lambda x: np.sum(x, axis=0),
        agg_fn=lambda vals, _: np.sum(vals, axis=0),
        remove_fn=lambda rands, _: -np.sum(rands, axis=0),
    ),
    "Mean": Metric(
        scale_rand=False,
        fit_fn=lambda x: np.mean(x, axis=0),
        agg_fn=lambda vals, weights: (weights.T @ vals) / np.sum(weights, axis=0),
        remove_fn=lambda rands, weights: -(weights.T @ rands) / np.sum(weights, axis=0),
    ),
    "Variance": Metric(
        # Calculating variance using sums should be numerically more stable
        scale_rand=False,
        fit_fn=lambda x: np.array(
            [
                np.sum(np.power(x, 2), axis=0),
                np.sum(x, axis=0),
            ]
        ),
        agg_fn=lambda vals, _: np.sum(vals, axis=0),
        remove_fn=lambda rands, _: -np.sum(rands, axis=0),
        output_fn=lambda vals, size: abs(
            (vals[0] - vals[1] ** 2 / sum(size)) / sum(size)
        ),
    ),
    "Std": Metric(
        # Calculating variance using sums should be numerically more stable
        scale_rand=False,
        fit_fn=lambda x: np.array(
            [
                np.sum(np.power(x, 2), axis=0),
                np.sum(x, axis=0),
            ]
        ),
        agg_fn=lambda vals, _: np.sum(vals, axis=0),
        remove_fn=lambda rands, _: -np.sum(rands, axis=0),
        output_fn=lambda vals, size: np.sqrt(
            np.abs((vals[0] - vals[1] ** 2 / sum(size)) / sum(size))
        ),
    ),
}


class SingleModel(AvgModel):
    """Model class for calculating single statistic implementing BaseModel."""

    model_type: str = "analytics"
    model: Dict[str, NDArray] = {
        "value": np.array([0]),
    }

    def __init__(self, data: dict):
        """Initialize model class from data dictionary.

        Args
            data: model loaded from JSON as dict
        """
        if data["model_name"] not in SUPPORTED_MODELS:
            raise Exception("Unsupported model type")

        self.model_name = data["model_name"]
        self.metric = SUPPORTED_MODELS[data["model_name"]]
        self.is_dirty = data.get("is_dirty", False)

        params = data.get("model_params", {})
        self._set_params(params)

        self.sample_size = data.get("sample_size", self.sample_size)
        if self.is_dirty:
            # Subtract random models (generated from seeds) from loaded model
            self.remove_noise_models(data.get("seeds", []))

    def _export_data(self) -> dict:
        """Get model data as dictionary for storing (and loading) model.

        Returns:
            dictionary containing model data which should be stored
        """
        return {
            "model_type": self.model_type,
            "model_name": self.model_name,
            "is_dirty": self.is_dirty,
            "model_params": self._get_params(),
            "sample_size": self.sample_size,
        }

    def get_random_models(
        self, seeds: List[int], _min: int = -100, _max: int = 100
    ) -> List[AvgModel]:
        """Generate models with random parameters.

        Args:
            seeds: list of seeds for randomness generation
            _min: minimum value of random number
            _max: maximum value of random number

        Returns:
            Returns copy of model with random variables.
        """
        assert len(seeds) == len(
            self.sample_size
        ), f"Can't generate random models. Num seeds ({len(seeds)}) and sizes ({len(self.sample_size)}) mismatch."

        models = []
        for seed, size in zip(seeds, self.sample_size):
            params = self._get_params()
            new_params = {}
            for param, array in params.items():
                self._set_seed(seed, param)

                new_params[param] = randomness.random_array_copy(array, _min, _max)
                if self.metric.scale_rand:
                    new_params[param] *= max(1, size)

            models.append(self.new_model(new_params))
        return models

    def remove_noise_models(self, seeds: List[int]) -> None:
        """Remove generate and remove random models from current model based on seeds.

        Args:
            seeds: list of seeds used for generating random models
        """
        if len(seeds) == 0 or not self.is_dirty:
            return

        noise_models = self.get_random_models(seeds)

        n_model, *other_models = noise_models
        n_model._agg_models_op(self.metric.remove_fn, other_models)

        self._agg_models_op(self.ops["sum_op"], [n_model])
        # Update sample size, because now we have clean aggregated model
        self.sample_size = [sum(self.sample_size)]
        self.is_dirty = False

    def _get_params(self) -> Dict[str, NDArray]:
        """Get dictionary of model parameters.

        Returns:
            dictionary of parameters as name to numpy array
        """
        return self.model

    def _set_params(self, params: Dict[str, NDArray]) -> None:
        """Set values of model parameters.

        Args:
            params: dictionary mapping from name of param to numpy array
        """
        self.model = {**self.model, **params}

    def _aggregate(self, models: List[AvgModel]) -> None:
        """Aggregation function on self + list of models.

        Args:
            models: list of models
        """
        if len(models) == 0:
            return
        self._agg_models_op(self.metric.agg_fn, models)

    def _fit(self, X: Any, y: Any) -> None:
        """Fit model on given data.

        Args:
            X: array like training data of shape (n_samples, n_features)
            y: array like target values of shape (n_samples,)
        """
        self.model["value"] = self.metric.fit_fn(y)

    def predict(self, X: Any) -> Any:
        """Use mode for prediction on given data.

        Args:
            X: array like data used for prediction of shape (n_samples, n_features)
        """
        print(
            f"{self.model_name} value is {self.metric.output_fn(self.model['value'], self.sample_size)}"
        )
        return self.metric.output_fn(self.model["value"], self.sample_size)


class Model(BaseModel):
    """Model class for single or multiple analytics."""

    model_type: str = "analytics"
    models = []

    def __init__(self, data: Union[dict, list]):
        """Initialize model class from data dictionary.

        Args
            data: model loaded from JSON as dict
        """
        data = data if isinstance(data, list) else [data]
        self.models = [SingleModel(d) for d in data]

    def _export_data(self) -> Union[dict, list]:
        """Get model data as dictionary for storing (and loading) model.

        Returns:
            dictionary containing model data which should be stored
        """
        if len(self.models) == 1:
            return self.models[0]._export_data()
        return [model._export_data() for model in self.models]

    def add_noise(self, seed: int) -> None:
        """Add pseudo random noise to the model.

        Args:
            seed: randomness seed to generate pseudo random model
        """
        # TODO: increment seed for each model (seed + i), need extra remove
        for model in self.models:
            model.add_noise(seed)

    def remove_noise_models(self, seeds: List[int]) -> None:
        """Remove generate and remove random models from current model based on seeds.

        Args:
            seeds: list of seeds used for generating random models
        """
        for model in self.models:
            model.remove_noise_models(seeds)

    def aggregate(self, models: List["Model"]) -> None:
        """Wrapper around aggregation function

        Args:
            models: list of models
        """
        for model, *other_models in zip(self.models, *[m.models for m in models]):
            model.aggregate(other_models)

    def fit(self, X: Any, y: Any) -> None:
        """Fit model on given data.

        Args:
            X: array like training data of shape (n_samples, n_features)
            y: array like target values of shape (n_samples,)
        """
        for model in self.models:
            model.fit(X, y)

    def predict(self, X: Any) -> Any:
        """Use mode for prediction on given data.

        Args:
            X: array like data used for prediction of shape (n_samples, n_features)
        """
        if len(self.models) == 1:
            return self.models[0].predict(X)
        return [model.predict(X) for model in self.models]



"""Testing loading of data."""
from pathlib import Path

import numpy as np
from sklearn.linear_model import LinearRegression

from feltlabs.config import parse_training_args
from feltlabs.core.data import load_data

# Test data both with and without header
csv_data = [
    "column_1,col_2,c3\n1,2,3\n0,2,3\n1,1,4\n0,2,5\n1,1,6",
    "1,2,3\n0,2,3\n1,1,4\n0,2,5\n1,1,6",
]


def test_csv_load(tmp_path: Path):
    # Create CSV files
    dataset_folder = tmp_path / "fake_did"
    dataset_folder.mkdir()
    for i in range(2):
        file_path = dataset_folder / f"{i}"
        with open(file_path, "w") as f:
            f.write(csv_data[i])

    args_str = f" --data_type csv --input_folder {tmp_path} --aggregation_key fe"
    # Define extra args with different output folder
    args = parse_training_args(args_str.split())
    X, y = load_data(args)

    assert len(X) == len(y), "Size of X and y are not matching"
    assert len(X) == 2 * 5, "Not all data loaded"
    assert np.all(X[:5] == np.array([[1, 2], [0, 2], [1, 1], [0, 2], [1, 1]]))
    assert np.all(y[:5] == np.array([3, 3, 4, 5, 6]))

    args_str = f" --data_type csv --target_column 0 --input_folder {tmp_path} --aggregation_key fe"
    # Define extra args with different output folder
    args = parse_training_args(args_str.split())
    X, y = load_data(args)
    assert len(X) == len(y), "Size of X and y are not matching"
    assert len(X) == 2 * 5, "Not all data loaded"
    assert np.all(X[:5] == np.array([[2, 3], [2, 3], [1, 4], [2, 5], [1, 6]]))
    assert np.all(y[:5] == np.array([1, 0, 1, 0, 1]))

    # Try if training works
    model = LinearRegression()
    model.fit(X, y)


# Test data both with and without header
data = [
    "column_1,col_2,c3\n1,2,3\n0,2,3\n1,1,4\n0,2,5\n1,1,6",
    "1,2,3\n0,2,3\n1,1,4\n0,2,5\n1,1,6",
    # Adding data config file:
    '{"name": "FELT_CONFIG", "data_type": "csv"}',
]


def test_data_config_load(tmp_path: Path):
    # Create CSV files
    dataset_folder = tmp_path / "fake_did"
    dataset_folder.mkdir()
    for i in range(len(data)):
        file_path = dataset_folder / f"{i}"
        with open(file_path, "w") as f:
            f.write(data[i])

    args_str = f" --data_type csv --input_folder {tmp_path} --aggregation_key fe"
    # Define extra args with different output folder
    args = parse_training_args(args_str.split())
    args.data_type = (
        "notInUse"  # Just test that data_type will be taken from config file
    )
    X, y = load_data(args)

    assert len(X) == len(y), "Size of X and y are not matching"
    assert len(X) == 2 * 5, "Not all data loaded"
    assert np.all(X[:5] == np.array([[1, 2], [0, 2], [1, 1], [0, 2], [1, 1]]))
    assert np.all(y[:5] == np.array([3, 3, 4, 5, 6]))

    args_str = f" --data_type csv --target_column 0 --input_folder {tmp_path} --aggregation_key fe"
    # Define extra args with different output folder
    args = parse_training_args(args_str.split())
    X, y = load_data(args)
    assert len(X) == len(y), "Size of X and y are not matching"
    assert len(X) == 2 * 5, "Not all data loaded"
    assert np.all(X[:5] == np.array([[2, 3], [2, 3], [1, 4], [2, 5], [1, 6]]))
    assert np.all(y[:5] == np.array([1, 0, 1, 0, 1]))

    # Try if training works
    model = LinearRegression()
    model.fit(X, y)



"""Module for storing and managing data files."""
from typing import Union

from feltlabs.core.cryptography import encrypt_nacl
from feltlabs.core.json_handler import json_load
from feltlabs.core.models import analytics_model, sklearn_model, tensorflow_model
from feltlabs.core.models.base_model import BaseModel
from feltlabs.typing import FileType


def _is_experimental(experimental: bool) -> None:
    """Raise exception if loading experimental model while experimental is False."""
    if not experimental:
        raise Exception("Loading experimental model while experimental is set False.")


def load_model(file: Union[FileType, dict], experimental: bool = True) -> BaseModel:
    """Load model from json file (intended for use in 3rd party programs).

    Args:
        file: path to json file containing model produced by FELT labs.
        experimental: allow loading experimental models (provides lower security)

    Returns:
        scikit-learn model
    """
    data = json_load(file)

    if data["model_type"] == "sklearn":
        return sklearn_model.Model(data)
    elif data["model_type"] == "analytics":
        return analytics_model.Model(data)
    elif data["model_type"] == "tensorflow":
        _is_experimental(experimental)
        return tensorflow_model.Model(data)
    raise Exception("Invalid model type.")


def encrypt_model(model: BaseModel, public_key: bytes) -> bytes:
    """Encrypt final model using provided public key.

    Args:
        model: scikit-learn trained model
        public_key: public key to use for encryption

    Returns:
        encrypted model as bytes
    """
    model_bytes = model.export_model()
    encrypted_model = encrypt_nacl(public_key, model_bytes)
    return encrypted_model



"""Module for executing model training."""
from typing import List, Optional

from feltlabs.config import TrainingConfig, parse_training_args
from feltlabs.core.data import load_data
from feltlabs.core.ocean import save_output
from feltlabs.core.storage import encrypt_model, load_model


def main(
    args_str: Optional[List[str]] = None,
    config: Optional[TrainingConfig] = None,
    output_name: str = "model",
):
    """Main function for executing from command line.

    Args:
        args_str: list with string arguments or None if using command line
        config: config object is used instead of arg parser if specified
        output_name: name of output model
    """
    if config is None:
        config = parse_training_args(args_str)
    # Load model
    model = load_model(config.custom_data_path, config.experimental)
    # Load data and train model
    X, y = load_data(config)
    model.fit(X, y)

    if config.solo_training:
        model = model.export_model()
    else:
        # Add randomness to model
        model.add_noise(config.seed)
        # Encrypt model using public key of aggregation
        model = encrypt_model(model, config.aggregation_key)

    # Save models into output
    save_output(output_name, model, config)
    print("Training finished.")
    return model


if __name__ == "__main__":
    main()



"""Testing model aggregation functions."""
import numpy as np

from feltlabs.core.models import analytics_model

X = [[0], [1], [2]]
y = [[0, 1, 2], [1, 2, 3], [0, 2, 4]]


def test_analytics_sum_aggregation():
    model_data = {"model_type": "analytics", "model_name": "Sum"}

    model1 = analytics_model.SingleModel(model_data)
    model1.fit(X, y[0])

    model2 = analytics_model.SingleModel(model_data)
    model2.fit(X, y[1])

    model3 = analytics_model.SingleModel(model_data)
    model3.fit(X, y[2])

    assert np.allclose(model1.predict(X), sum(y[0]))
    assert np.allclose(model2.predict(X), sum(y[1]))

    agg_model = model1.new_model()
    agg_model.aggregate([model2])
    assert np.allclose(agg_model.predict(X), sum(y[0] + y[1]))

    agg_model = model1.new_model()
    agg_model.aggregate([model3])
    assert np.allclose(agg_model.predict(X), sum(y[0] + y[2]))

    enc_model1 = model1.new_model()
    enc_model1.add_noise(10)
    assert not np.allclose(enc_model1.predict(X), sum(y[0]))

    enc_model2 = model2.new_model()
    enc_model2.add_noise(11)
    assert not np.allclose(enc_model2.predict(X), sum(y[1]))

    enc_model3 = model3.new_model()
    enc_model3.add_noise(12)
    assert not np.allclose(enc_model3.predict(X), sum(y[2]))

    final_model = enc_model1
    final_model.aggregate([enc_model2, enc_model3])
    final_model.remove_noise_models([10, 11, 12])
    assert np.allclose(final_model.predict(X), sum(y[0] + y[1] + y[2]))


def test_analytics_mean_aggregation():
    model_data = {"model_type": "analytics", "model_name": "Mean"}

    model1 = analytics_model.SingleModel(model_data)
    model1.fit(X, y[0])

    model2 = analytics_model.SingleModel(model_data)
    model2.fit(X, y[1])

    model3 = analytics_model.SingleModel(model_data)
    model3.fit(X, y[2])

    assert np.allclose(model1.predict(X), np.mean(y[0]))
    assert np.allclose(model2.predict(X), np.mean(y[1]))

    agg_model = model1.new_model()
    agg_model.aggregate([model2])
    assert np.allclose(agg_model.predict(X), np.mean(y[0] + y[1]))

    agg_model = model1.new_model()
    agg_model.aggregate([model3])
    assert np.allclose(agg_model.predict(X), np.mean(y[0] + y[2]))

    enc_model1 = model1.new_model()
    enc_model1.add_noise(10)
    assert not np.allclose(enc_model1.predict(X), np.mean(y[0]))

    enc_model2 = model2.new_model()
    enc_model2.add_noise(11)
    assert not np.allclose(enc_model2.predict(X), np.mean(y[1]))

    enc_model3 = model3.new_model()
    enc_model3.add_noise(12)
    assert not np.allclose(enc_model3.predict(X), np.mean(y[2]))

    final_model = enc_model1
    final_model.aggregate([enc_model2, enc_model3])
    final_model.remove_noise_models([10, 11, 12])
    assert np.allclose(final_model.predict(X), np.mean(y[0] + y[1] + y[2]))


def test_analytics_variance_aggregation():
    model_data = {"model_type": "analytics", "model_name": "Variance"}

    model1 = analytics_model.SingleModel(model_data)
    model1.fit(X, y[0])

    model2 = analytics_model.SingleModel(model_data)
    model2.fit(X, y[1])

    model3 = analytics_model.SingleModel(model_data)
    model3.fit(X, y[2])

    assert np.allclose(model1.predict(X), np.var(y[0]))
    assert np.allclose(model2.predict(X), np.var(y[1]))

    agg_model = model1.new_model()
    agg_model.aggregate([model2])
    assert np.allclose(agg_model.predict(X), np.var(y[0] + y[1]))

    agg_model = model1.new_model()
    agg_model.aggregate([model3])
    assert np.allclose(agg_model.predict(X), np.var(y[0] + y[2]))

    enc_model1 = model1.new_model()
    enc_model1.add_noise(10)
    assert not np.allclose(enc_model1.predict(X), np.var(y[0]))

    enc_model2 = model2.new_model()
    enc_model2.add_noise(11)
    assert not np.allclose(enc_model2.predict(X), np.var(y[1]))

    enc_model3 = model3.new_model()
    enc_model3.add_noise(12)
    assert not np.allclose(enc_model3.predict(X), np.var(y[2]))

    final_model = enc_model1
    final_model.aggregate([enc_model2, enc_model3])
    final_model.remove_noise_models([10, 11, 12])
    assert np.allclose(final_model.predict(X), np.var(y[0] + y[1] + y[2]))


def test_analytics_std_aggregation():
    model_data = {"model_type": "analytics", "model_name": "Std"}

    model1 = analytics_model.SingleModel(model_data)
    model1.fit(X, y[0])

    model2 = analytics_model.SingleModel(model_data)
    model2.fit(X, y[1])

    model3 = analytics_model.SingleModel(model_data)
    model3.fit(X, y[2])

    assert np.allclose(model1.predict(X), np.std(y[0]))
    assert np.allclose(model2.predict(X), np.std(y[1]))

    agg_model = model1.new_model()
    agg_model.aggregate([model2])
    assert np.allclose(agg_model.predict(X), np.std(y[0] + y[1]))

    agg_model = model1.new_model()
    agg_model.aggregate([model3])
    assert np.allclose(agg_model.predict(X), np.std(y[0] + y[2]))

    enc_model1 = model1.new_model()
    enc_model1.add_noise(10)
    assert not np.allclose(enc_model1.predict(X), np.std(y[0]))

    enc_model2 = model2.new_model()
    enc_model2.add_noise(11)
    assert not np.allclose(enc_model2.predict(X), np.std(y[1]))

    enc_model3 = model3.new_model()
    enc_model3.add_noise(12)
    assert not np.allclose(enc_model3.predict(X), np.std(y[2]))

    final_model = enc_model1
    final_model.aggregate([enc_model2, enc_model3])
    final_model.remove_noise_models([10, 11, 12])
    assert np.allclose(final_model.predict(X), np.std(y[0] + y[1] + y[2]))



"""Module providing model class interface."""
import copy
import hashlib
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Optional

import numpy as np
from numpy.typing import NDArray

from feltlabs.core import randomness
from feltlabs.core.json_handler import json_dump
from feltlabs.typing import PathType


class BaseModel(ABC):
    """Base model class for federated learning.
    Each model must subclass BaseModel so that necessary functions for federated
    learning are implemented.
    """

    model_type: str
    model_name: str
    sample_size: List[int] = [0]
    is_dirty: bool  # True if randomness was added

    @abstractmethod
    def __init__(self, data: dict):
        """Initialize model class from data dictionary.

        Args
            data: model loaded from JSON as dict
        """

    def export_model(self, filename: Optional[PathType] = None) -> bytes:
        """Export model to bytes and optionally store it as JSON file.

        Args:
            filename: path to exported file

        Returns:
            bytes of JSON file
        """
        data = self._export_data()
        data_bytes = json_dump(data)
        if filename:
            with open(filename, "wb") as f:
                f.write(data_bytes)

        return data_bytes

    def _set_seed(self, seed: int, param: str) -> None:
        """Set randomness seed of given parameter.

        Args:
            seed: seed provided for the whole training
            param: parameter for which we are setting the seed
        """
        seed = int(
            hashlib.sha256(bytes(f"{seed};{param}", "utf-8")).hexdigest(), 16
        ) % (2**32 - 1)
        randomness.set_seed(seed)

    # TODO: Specify typing for fit and predict
    @abstractmethod
    def fit(self, X: Any, y: Any) -> None:
        """Wrapper around model fitting.

        Args:
            X: array like training data of shape (n_samples, n_features)
            y: array like target values of shape (n_samples,)
        """

    @abstractmethod
    def aggregate(self, models: List["BaseModel"]) -> None:
        """Wrapper around aggregation function

        Args:
            models: list of models
        """

    @abstractmethod
    def predict(self, X: Any) -> Any:
        """Use mode for prediction on given data.

        Args:
            X: array like data used for prediction of shape (n_samples, n_features)
        """

    @abstractmethod
    def add_noise(self, seed: int) -> None:
        """Add pseudo random noise to the model.

        Args:
            seed: randomness seed to generate pseudo random model
        """

    @abstractmethod
    def remove_noise_models(self, seeds: List[int]) -> None:
        """Remove generate and remove random models from current model based on seeds.

        Args:
            seeds: list of seeds used for generating random models
        """

    @abstractmethod
    def _export_data(self) -> dict:
        """Get model data as dictionary for storing (and loading) model.

        Returns:
            dictionary containing model data which should be stored
        """


class AvgModel(BaseModel):
    """Model based on Based Model necessary for basic averaging federated learning.
    Most of the models should subclass from this model.
    """

    ops = {
        "sum_op": lambda x, _: np.sum(x, axis=0),
        "mean_op": lambda x, _: np.mean(x, axis=0),
    }

    def new_model(self, params: Dict[str, NDArray] = {}) -> "AvgModel":
        """Create copy of model and set new parameters.

        Args:
            params: values to set to the new model
        """
        new_model = copy.deepcopy(self)
        new_model._set_params(params)
        return new_model

    def add_noise(self, seed: int) -> None:
        """Add pseudo random noise to the model.

        Args:
            seed: randomness seed to generate pseudo random model
        """
        assert len(self.sample_size) == 1, "Can't add randomness to aggregated model."
        rand_model = self.get_random_models([seed])
        self._agg_models_op(self.ops["sum_op"], rand_model)
        self.is_dirty = True

    def get_random_models(
        self, seeds: List[int], _min: int = -100, _max: int = 100
    ) -> List["AvgModel"]:
        """Generate models with random parameters.

        Args:
            seeds: list of seeds for randomness generation
            _min: minimum value of random number
            _max: maximum value of random number

        Returns:
            Returns copy of model with random variables.
        """
        assert len(seeds) == len(
            self.sample_size
        ), f"Can't generate random models. Num seeds ({len(seeds)}) and sizes ({len(self.sample_size)}) mismatch."

        models = []
        # TODO: Right now we are not using "size" for the sklearn models
        for seed, size in zip(seeds, self.sample_size):
            params = self._get_params()
            new_params = {}
            for param, array in params.items():
                self._set_seed(seed, param)

                if type(array) == list:
                    value = [randomness.random_array_copy(a, _min, _max) for a in array]
                else:
                    value = randomness.random_array_copy(array, _min, _max)

                new_params[param] = value

            models.append(self.new_model(new_params))
        return models

    # TODO: Specify typing for fit and predict
    def fit(self, X: Any, y: Any) -> None:
        """Wrapper around model fitting.

        Args:
            X: array like training data of shape (n_samples, n_features)
            y: array like target values of shape (n_samples,)
        """
        self.sample_size = [len(y)]
        self._fit(X, y)

    def aggregate(self, models: List["AvgModel"]) -> None:
        """Wrapper around aggregation function

        Args:
            models: list of models
        """
        self.sample_size.extend([m.sample_size[0] for m in models])
        self._aggregate(models)

    def export_model(self, filename: Optional[PathType] = None) -> bytes:
        """Export model to bytes and optionally store it as JSON file.

        Args:
            filename: path to exported file

        Returns:
            bytes of JSON file
        """
        data = self._export_data()
        data_bytes = json_dump(data)
        if filename:
            with open(filename, "wb") as f:
                f.write(data_bytes)

        return data_bytes

    def _agg_models_op(self, op: Callable, models: List["AvgModel"]) -> None:
        """Perform aggregation operation on list of models.

        Args:
            op: function to run on on values, with definition fn(model_values, weights)
            models: list of models
        """
        models_params = [m._get_params() for m in [self, *models]]
        models_weights = np.array([m.sample_size[0] for m in [self, *models]])

        new_params = {}
        for param in models_params[0]:
            if not all(param in params for params in models_params):
                print(f"WARNING: Parameter '{param}' is not present in all models.")
                continue

            is_list = isinstance(models_params[0][param], list)

            values = [params[param] for params in models_params]
            # Convert to array of objects to prevent warning (elements have different shape)
            values = np.array(values, dtype=object) if is_list else values
            val = op(values, models_weights)

            if is_list:
                val = op(values, models_weights)
                new_params[param] = list(val)
            else:
                val = op(values, models_weights)
                new_params[param] = val

        self._set_params(new_params)

    @abstractmethod
    def _aggregate(self, models: List["AvgModel"]) -> None:
        """Aggregation function on self + list of models, specific for given model.

        Args:
            models: list of models
        """

    @abstractmethod
    def _fit(self, X: Any, y: Any) -> None:
        """Fit model on given data, specific for given model type.

        Args:
            X: array like training data of shape (n_samples, n_features)
            y: array like target values of shape (n_samples,)
        """

    @abstractmethod
    def _get_params(self) -> Dict[str, NDArray]:
        """Get dictionary of model parameters.

        Returns:
            dictionary of parameters as name to numpy array
        """

    @abstractmethod
    def _set_params(self, params: Dict[str, NDArray]) -> None:
        """Set values of model parameters.

        Args:
            params: dictionary mapping from name of param to numpy array
        """



"""Function specific for Ocean's compute-to-data environment."""
import json
from pathlib import Path
from typing import Dict, List, Tuple

from feltlabs.config import OceanConfig


class DatasetConfig:
    """Config class for loading datasets."""

    data_type: str
    load_code: str
    _files: Dict[int, Path]

    @property
    def files(self) -> List[Tuple[Path, int]]:
        """Get files as list sorted by index."""
        return sorted(
            [(path, index) for index, path in self._files.items()], key=lambda x: x[1]
        )

    def __init__(self):
        """Init config class with default values."""
        self.data_type = "default"
        self._files = {}
        self.load_code = ""

    def parse_config(self, path: Path) -> None:
        """Load FELT config file from path."""
        with path.open() as f:
            config = json.load(f)

        self.data_type = config.get("data_type", self.data_type)
        self.load_code = config.get("load_code", self.load_code)

    def add_file(self, path: Path) -> None:
        """Add data file path to dataset object."""
        index = int(path.name)
        self._files[index] = path


def _is_dataset_config(path: Path) -> bool:
    """Check if file is FELT dataset config file."""
    try:
        with path.open() as f:
            config = json.load(f)
            if config["name"] == "FELT_CONFIG":
                return True
    except Exception:
        pass
    return False


def get_datasets(config: OceanConfig) -> Dict[str, DatasetConfig]:
    """Get all dataset paths provided in Ocean's compute job environment.

    Args:
        config: ocean config containing output path

    Returns:
        dictionary mapping dataset DID to dataset loading config
    """
    did_folders = [(p, p.name) for p in config.input_folder.iterdir() if p.is_dir()]

    datasets = {}
    for folder, did in did_folders:
        datasets = {did: DatasetConfig()}

        for path in folder.glob("**/*"):
            if path.is_file() and _is_dataset_config(path):
                datasets[did].parse_config(path)
            elif path.is_file():
                datasets[did].add_file(path)

    return datasets


def save_output(name: str, data: bytes, config: OceanConfig) -> None:
    """Save data to output folder by given name.

    Args:
        name: name of file to write
        data: data as bytes to store in the file
        config: ocean config containing output path
    """
    with open(config.output_folder / name, "wb+") as f:
        f.write(data)



"""Testing model aggregation functions."""
import numpy as np

from feltlabs.core.models import sklearn_model

X = [[0], [1], [2]]
y = [[0, 1, 2], [1, 2, 3], [0, 2, 4]]


def test_sklearn_linreg_aggregation():
    model_data = {"model_type": "sklearn", "model_name": "LinearRegression"}

    model1 = sklearn_model.Model(model_data)
    model1.fit(X, y[0])

    model2 = sklearn_model.Model(model_data)
    model2.fit(X, y[1])

    model3 = sklearn_model.Model(model_data)
    model3.fit(X, y[2])

    assert np.allclose(model1.predict(X), y[0])
    assert np.allclose(model2.predict(X), y[1])

    agg_model = model1.new_model()
    agg_model.aggregate([model2])
    assert np.allclose(agg_model.predict(X), [0.5, 1.5, 2.5])

    agg_model = model1.new_model()
    agg_model.aggregate([model3])
    assert np.allclose(agg_model.predict(X), [0, 1.5, 3])

    enc_model1 = model1.new_model()
    enc_model1.add_noise(10)
    assert not np.allclose(enc_model1.predict(X), y[0])

    enc_model2 = model2.new_model()
    enc_model2.add_noise(11)
    assert not np.allclose(enc_model2.predict(X), y[1])

    final_model = enc_model1
    final_model.aggregate([enc_model2])
    final_model.remove_noise_models([10, 11])
    assert np.allclose(final_model.predict(X), [0.5, 1.5, 2.5])


def test_sklearn_lasso_aggregation():
    model_data = {
        "model_type": "sklearn",
        "model_name": "Lasso",
        "init_params": {"alpha": 0.05},
    }

    model1 = sklearn_model.Model(model_data)
    model1.fit(X, y[0])

    model2 = sklearn_model.Model(model_data)
    model2.fit(X, y[1])

    model3 = sklearn_model.Model(model_data)
    model3.fit(X, y[2])

    assert np.allclose(model1.predict(X), y[0], atol=0.20)
    assert np.allclose(model2.predict(X), y[1], atol=0.20)

    agg_model = model1.new_model()
    agg_model.aggregate([model2])
    assert np.allclose(agg_model.predict(X), [0.5, 1.5, 2.5], atol=0.20)

    agg_model = model1.new_model()
    agg_model.aggregate([model3])
    assert np.allclose(agg_model.predict(X), [0, 1.5, 3], atol=0.20)

    enc_model1 = model1.new_model()
    enc_model1.add_noise(10)
    assert not np.allclose(enc_model1.predict(X), y[0], atol=0.20)

    enc_model2 = model2.new_model()
    enc_model2.add_noise(11)
    assert not np.allclose(enc_model2.predict(X), y[1], atol=0.20)

    final_model = enc_model1
    final_model.aggregate([enc_model2])
    final_model.remove_noise_models([10, 11])
    assert np.allclose(final_model.predict(X), [0.5, 1.5, 2.5], atol=0.20)


def test_sklearn_ridge_aggregation():
    model_data = {
        "model_type": "sklearn",
        "model_name": "Ridge",
        "init_params": {"alpha": 0.05},
    }

    model1 = sklearn_model.Model(model_data)
    model1.fit(X, y[0])

    model2 = sklearn_model.Model(model_data)
    model2.fit(X, y[1])

    model3 = sklearn_model.Model(model_data)
    model3.fit(X, y[2])

    assert np.allclose(model1.predict(X), y[0], atol=0.20)
    assert np.allclose(model2.predict(X), y[1], atol=0.20)

    agg_model = model1.new_model()
    agg_model.aggregate([model2])
    assert np.allclose(agg_model.predict(X), [0.5, 1.5, 2.5], atol=0.20)

    agg_model = model1.new_model()
    agg_model.aggregate([model3])
    assert np.allclose(agg_model.predict(X), [0, 1.5, 3], atol=0.20)

    enc_model1 = model1.new_model()
    enc_model1.add_noise(10)
    assert not np.allclose(enc_model1.predict(X), y[0], atol=0.20)

    enc_model2 = model2.new_model()
    enc_model2.add_noise(11)
    assert not np.allclose(enc_model2.predict(X), y[1], atol=0.20)

    final_model = enc_model1
    final_model.aggregate([enc_model2])
    final_model.remove_noise_models([10, 11])
    assert np.allclose(final_model.predict(X), [0.5, 1.5, 2.5], atol=0.20)



"""Module for aggregating outputs of local training."""
from typing import List, Optional

from feltlabs.config import AggregationConfig, parse_aggregation_args
from feltlabs.core.data import load_models
from feltlabs.core.ocean import save_output
from feltlabs.core.storage import encrypt_model


def main(
    args_str: Optional[List[str]] = None,
    config: Optional[AggregationConfig] = None,
    output_name: str = "model",
):
    """Main function for executing from command line.

    Args:
        args_str: list with string arguments or None if using command line
        config: config object is used instead of arg parser if specified
        output_name: name of output model
    """
    if config is None:
        config = parse_aggregation_args(args_str)
    # Load models
    models = load_models(config)
    if len(models) < config.min_models:
        raise Exception(
            f"Not enough models for aggregation, loaded {len(models)} models (required {config.min_models})."
        )
    # Aggregate
    model, *other_models = models
    model.aggregate(other_models)
    # Encrypt final model using scientist public key if provided
    if config.public_key:
        model = encrypt_model(model, config.public_key)
    else:
        model = model.export_model()
    # Save model (bytes) into output
    save_output(output_name, model, config)
    print("Aggregation finished.")
    return model


if __name__ == "__main__":
    main()



"""Module for generating random numbers used by models."""
import random
from typing import Union

import numpy as np
from numpy.typing import NDArray


def set_seed(seed: int) -> None:
    """Set seed of random generator."""
    np.random.seed(seed)
    random.seed(seed)


def random_array_copy(
    array: Union[NDArray, int, float], min: float, max: float
) -> Union[NDArray, float]:
    """Create random array with same shape and data type as original array.

    Args:
        array: original array based on witch we want to create new array
        min: minumal value in new array
        max: maximum value in new array

    Returns:
        New random array of same shape and type as original array
    """
    if isinstance(array, np.ndarray):
        return (max - min) * np.random.random(array.shape) + min
    return (max - min) * np.random.random() + min



"""Module for handling sklearn models."""
from typing import Any, Dict, List

import numpy as np
from numpy.typing import NDArray
from sklearn import linear_model, neighbors, neural_network, preprocessing

from feltlabs.core.models.base_model import AvgModel

# TODO: SVM attributes  ["dual_coef_", "support_", "support_vectors_", "_n_support"
# Attributes and data type casting for them (done only after removing randomness)
ATTRIBUTE_LIST = {
    "coef_": None,
    "intercept_": None,
    "coefs_": None,
    "intercepts_": None,
    "classes_": lambda x: np.rint(x).astype(np.uint),
    "centroids_": None,
    "n_iter_": np.rint,
    "n_layers_": round,
    "n_outputs_": round,
    # Needed by MLPClassifier
    "t_": round,
    "loss_curve_": None,
    "best_loss_": None,
}
FIXED_ATTRIBUTE_LIST = [
    "out_activation_",
    "_no_improvement_count",
]

SUPPORTED_MODELS = {
    # Regression
    "LinearRegression": linear_model.LinearRegression,
    "Lasso": linear_model.Lasso,
    "Ridge": linear_model.Ridge,
    "ElasticNet": linear_model.ElasticNet,
    "LassoLars": linear_model.LassoLars,
    # Classification
    "LogisticRegression": linear_model.LogisticRegression,
    "SGDClassifier": linear_model.SGDClassifier,
    # Clustering
    "NearestCentroidClassifier": neighbors.NearestCentroid,
    # Neural Networks
    # TODO: Limit size of hidden layers
    "MLPClassifier": neural_network.MLPClassifier,
    "MLPRegressor": neural_network.MLPRegressor,
}


class Model(AvgModel):
    """Model class for scikit-learn models implementing BaseModel."""

    model_type: str = "sklearn"

    def __init__(self, data: dict):
        """Initialize model class from data dictionary.

        Args
            data: model loaded from JSON as dict
        """
        if data["model_name"] not in SUPPORTED_MODELS:
            raise Exception("Unsupported model type")

        self.model_name = data["model_name"]
        self.is_dirty = data.get("is_dirty", False)

        model_class = SUPPORTED_MODELS[self.model_name]
        self.model = model_class(**data.get("init_params", {}))

        params = data.get("model_params", {})
        self._set_params(params)

        self.sample_size = data.get("sample_size", self.sample_size)

        if self.is_dirty:
            # Subtract random models (generated from seeds) from loaded model
            self.remove_noise_models(data.get("seeds", []))
        else:
            self._init_post_clean()

    def _init_post_clean(self):
        """Run extra initialization for clean model."""
        # Bit hacky solution adding label binarizer for MLPClassifier
        if (
            self.model_name == "MLPClassifier"
            and not hasattr(self.model, "_label_binarizer")
            and hasattr(self.model, "classes_")
        ):
            self.model._label_binarizer = preprocessing.LabelBinarizer()
            self.model._label_binarizer.fit(self.model.classes_)

    def _export_data(self) -> dict:
        """Get model data as dictionary for storing (and loading) model.

        Returns:
            dictionary containing model data which should be stored
        """
        return {
            "model_type": self.model_type,
            "model_name": self.model_name,
            "is_dirty": self.is_dirty,
            "init_params": self.model.get_params(),  # Get params of sklearn models
            "model_params": {
                **self._get_params(),
                **self._get_params(FIXED_ATTRIBUTE_LIST),
            },
            "sample_size": self.sample_size,
        }

    def remove_noise_models(self, seeds: List[int]) -> None:
        """Remove generate and remove random models from current model based on seeds.

        Args:
            seeds: list of seeds used for generating random models
        """
        if len(seeds) == 0 or not self.is_dirty:
            return

        noise_models = self.get_random_models(seeds)

        op = lambda x, _: -1 * np.mean(x, axis=0)

        n_model, *other_models = noise_models
        n_model._agg_models_op(op, other_models)

        self._agg_models_op(self.ops["sum_op"], [n_model])
        # Cast parameters to required types
        self._set_params(self._get_params(), type_cast=True)
        # Update sample size, because now we have clean aggregated model
        self.sample_size = [sum(self.sample_size)]
        self.is_dirty = False
        # Init other params based on clean model
        self._init_post_clean()

    def _get_params(
        self, attributes: list = list(ATTRIBUTE_LIST)
    ) -> Dict[str, NDArray]:
        """Get dictionary of model parameters.

        Args:
            attributes: list of attributes to get from the object

        Returns:
            dictionary of parameters as name to numpy array
        """
        params = {}
        for p in attributes:
            if hasattr(self.model, p) and getattr(self.model, p) is not None:
                params[p] = getattr(self.model, p)
        return params

    def _set_params(self, params: Dict[str, NDArray], type_cast: bool = False) -> None:
        """Set values of model parameters.

        Args:
            params: dictionary mapping from name of param to numpy array
            type_cast: set true if types should be cast to expected type
        """
        for param, value in params.items():
            if type_cast and ATTRIBUTE_LIST[param] is not None:
                value = ATTRIBUTE_LIST[param](value)
            setattr(self.model, param, value)

    def _aggregate(self, models: List[AvgModel]) -> None:
        """Aggregation function on self + list of models.

        Args:
            models: list of models
        """
        if len(models) == 0:
            return
        self._agg_models_op(self.ops["mean_op"], models)

    def _fit(self, X: Any, y: Any) -> None:
        """Fit model on given data.

        Args:
            X: array like training data of shape (n_samples, n_features)
            y: array like target values of shape (n_samples,)
        """
        # TODO: Find best way to set best_loss_
        # Adjust best loss for the early stopping
        if hasattr(self.model, "best_loss_"):
            self.model.best_loss_ *= 2

        # Reset early stopping and max_training per round
        reset_attr = ["_no_improvement_count", "n_iter_"]
        for attr in reset_attr:
            if hasattr(self.model, attr):
                setattr(self.model, attr, 0)

        self.model.fit(X, y)

    def predict(self, X: Any) -> Any:
        """Use mode for prediction on given data.

        Args:
            X: array like data used for prediction of shape (n_samples, n_features)
        """
        return self.model.predict(X)



"""Test training process."""
from pathlib import Path

import pytest
from nacl.public import PrivateKey

from feltlabs.algorithm import aggregate, train
from feltlabs.config import parse_training_args
from feltlabs.core.data import load_data
from feltlabs.core.json_handler import json_dump
from feltlabs.core.storage import load_model

# data = np.array([[0, 0], [1, 1], [2, 2]]), np.array([0, 1, 2])
# Right now it uses test dataset

aggregation_key = PrivateKey.generate()
scientist_key = PrivateKey.generate()

model_def = {
    "model_type": "sklearn",
    "model_name": "LinearRegression",
}


def test_training(tmp_path: Path):
    input_folder = tmp_path / "input"
    output_folder = tmp_path / "output" / "fake_did"
    output_folder2 = tmp_path / "output2"

    input_folder.mkdir()
    output_folder.mkdir(parents=True)
    output_folder2.mkdir()

    # Create custom data file (containing model definition)
    with open(input_folder / "algoCustomData.json", "wb") as f:
        f.write(json_dump(model_def))

    enc_models, seeds = [], []

    ### Training section ###
    args_str = f"--output_folder {output_folder}"
    args_str += f" --input_folder {input_folder}"
    args_str += f" --aggregation_key {bytes(aggregation_key.public_key).hex()}"

    # Define extra args with different output folder
    args = parse_training_args(args_str.split())

    for i in range(2):
        args_str_final = f"{args_str} --seed {i}"
        enc_model = train.main(args_str_final.split(), output_name=f"{i}")

        enc_models.append(enc_model)
        seeds.append(i)

    ### Aggregation section ###
    args_str = f"--output_folder {output_folder2}"
    args_str += f" --input_folder {output_folder.parent}"
    args_str += f" --private_key {bytes(aggregation_key).hex()}"

    enc_final_model = aggregate.main(args_str.split(), output_name="final_model")

    ### Test final results ###
    final_model = load_model(enc_final_model)
    final_model.remove_noise_models(seeds)

    # Predict
    data = load_data(args)
    final_model.predict(data[0])

    ### Aggregation section ###
    args_str = f"--output_folder {output_folder2}"
    args_str += f" --input_folder {output_folder}"
    args_str += f" --private_key {bytes(aggregation_key).hex()}"
    args_str += f" --min_models 3"

    with pytest.raises(Exception):
        # Should fail because at least 3 models are required
        enc_final_model = aggregate.main(args_str.split(), output_name="final_model")
